-- Mute, Unmute, or Ramp Up/Down Volume on Scene Transition
--!rename on create/load/delete of audio source
--!convert to lua native logging
--!simplify rename logic, based on rename signal working
--!show scene name(s) in show_source_details diagnostic 
--!add support for audio source subsets, for example "active" or "entering/leaving"
--!add support for transition override duration

bit = require("bit")
--global constants 
obs = obslua
version = "v1.1 by DQM 3/26/2025"
support = "dennis@musefeed.net"
sourceType="Audio Transition"
all="All"
none="None"
entering="entering"
leaving="leaving"	

--global variables
transitionSignalsEnabled=false
eventLogEnabled=false	
faderLogEnabled=false
errorPopupEnabled=false
transitionsEnabled=true
enterSceneName=" "
leaveSceneName=" "
loadAttempts=0

--script properties 
faderDelayOffset= 0  	--default step delay in milliseconds
faderDelayBaseline = 60 -- delaybaseline+delaymin must be > 0
faderDelayMin=-50   --milliseconds
faderDelayMax=50   	--milliseconds
faderDurationMax = 10  	--default duration max seconds
faderMin = -60
faderMax = 0
vs=8  --volume significant digits

--global tables
faderData={}  	-- used to track fader progress
leaveData={}  	-- used to save audio transitions for leaving scenes
enterData={}	-- used to save audio transitions for entering scenes
audioData={}    -- used to save audio state for restore actions 
signalData={}	-- used to save transition signals
function script_description()
	return [[
<div style='text-align: justify'>
<div style='color: cyan; text-align: center; font-weight: bold'>]]..sourceType..[[</div>
<div style='text-align: center; color: white'>]]..version..[[</div>
<div style='text-align: center; color: white'>]]..support..[[</div>
<div style='line-height:50%;'>&nbsp;</div><div style='line-height: normal'></div>
<div style='color: yellow; line-height:100%'>Mutes, Unmutes, or changes the volume of an audio source during a scene transition.
<div style='line-height:50%;'>&nbsp;</div><div style='line-height: normal'></div>
'Mute' and 'Unmute' are self-explanatory. 'Fade' adjusts the volume up or down in gradual steps (after unmuting).
 'Ramp' fades after silencing, first.
 'Set' adjusts the volume without the gradual fade.
 'Save' captures the volume and mute state for a subsequent restore.
 'Restore' returns the audio source to the most recently saved volume and mute state.
<div style='line-height:50%;'>&nbsp;</div><div style='line-height: normal'></div>
Audio Transitions <u>leaving</u> the Program Scene are initiated before those <u>entering</u> it.
 If more then one ramp or fade operation is encountered for the same Audio Source,
 each replaces the one preceeding it, in succession.  
<div style='line-height:50%;'>&nbsp;</div><div style='line-height: normal'></div>
</div>
<hr>
<div>
<div style='line-height:50%;'>&nbsp;</div><div style='line-height: normal'></div>
Fader Delay Offset adjusts the timing between fader steps. Smaller delays result in more,
 smaller steps and a smoother transition. However, steps add overhead and too many can make
 fading exceed the intended duration.
</div>
</div>
]]
end
function script_properties()
	local msg=""
	local props = obs.obs_properties_create()
	scriptProp=obs.obs_properties_add_bool(props, "enable_transitions", "Enable Audio Transitions")
	obs.obs_property_set_enabled(scriptProp,true)
	msg="Allow this script to transition audio sources"
	obs.obs_property_set_long_description(scriptProp,msg)
	
	local scriptProp=obs.obs_properties_add_int_slider(props, "fader_delay_offset", "Fader Delay Offset (from ".. faderDelayBaseline.. " milliseconds) ", faderDelayMin, faderDelayMax, 1)
	msg="Adjust fader delay to compensate for system performance differences"
	obs.obs_property_set_long_description(scriptProp,msg)
	scriptProp=obs.obs_properties_add_bool(props, "error_popup", "Enable Error Log Popup")
	obs.obs_property_set_enabled(scriptProp,true)
	msg="Open the script log when an error is posted\nErrors are recorded in the log, regardless"
	obs.obs_property_set_long_description(scriptProp,msg)
	scriptProp=obs.obs_properties_add_bool(props, "event_log", "Enable Event Logging")
	obs.obs_property_set_enabled(scriptProp,true)
	msg="Issue diagnostic messages showing transition events"
	obs.obs_property_set_long_description(scriptProp,msg) 	
	scriptProp=obs.obs_properties_add_bool(props, "fader_log", "Enable Fader Logging")
	obs.obs_property_set_enabled(scriptProp,true)
	msg="Issue diagnostic messages showing fader progress"
	obs.obs_property_set_long_description(scriptProp,msg)
	scriptProp=obs.obs_properties_add_button(props,"diagnostics_button","Show Sources and Volume Conversions",showDiagnostics)
	obs.obs_property_set_enabled(scriptProp, true)
	obs.obs_property_set_long_description(scriptProp,"Show audio source information in script log") 
	return props
end 
function script_defaults(settings)
	obs.obs_data_set_default_int(settings, "fader_delay_offset",0)
	obs.obs_data_set_default_bool(settings, "fader_log",false)
	obs.obs_data_set_default_bool(settings, "event_log",false)
	obs.obs_data_set_default_bool(settings, "error_popup",false)
	obs.obs_data_set_default_bool(settings, "enable_transitions",true)
end
function script_load(settings)
	print (sourceType.." "..version)
	registerPlugin()
	obs.obs_frontend_add_event_callback(onFrontendEvent)
	--need to wrap up initiation after the current scene is loaded.  However, can't use the --on_frontend_event_finished_loading signal because it does not fire on a refresh.
	--this invokes a callback every half-second to check for the presence of current_scene so that
	--initiation can completed.
	obs.timer_add(onScriptLoaded,500)
end
function script_update(settings)
	faderDelayOffset = obs.obs_data_get_int(settings, "fader_delay_offset")
	faderLogEnabled = obs.obs_data_get_bool(settings, "fader_log")
	eventLogEnabled = obs.obs_data_get_bool(settings, "event_log")
	errorPopupEnabled= obs.obs_data_get_bool(settings, "error_popup")
	transitionsEnabled = obs.obs_data_get_bool(settings,"enable_transitions")
end
function script_unload(settings)
--script cleanup
--signal handler are managed automatically according to doc
	obs.obs_frontend_remove_event_callback(onFrontendEvent)
	elog(numerate(nil, sourceType.." Unloaded"))
end

function source_create(settings, source)
		local sourceName = obs.obs_source_get_name(source)  
		local sourceNo = obs.obs_data_get_string(settings, "source_no")
		elog (nil,"Activating "..sourceType..numerate(sourceNo).." as ["..sourceName.."]")
		local userdata = {}
		local sh = obs.obs_source_get_signal_handler(source)		
		obs.signal_handler_connect(sh,"save",onSourceSave) --Save callback
		obs.signal_handler_connect(sh,"rename",onSourceRename) --rename callback
		return userdata  
	end
function source_properties(userdata)
	props = obs.obs_properties_create()
	local sourceProp =obs.obs_properties_add_text(props,"source_type","Source Type",obs.OBS_TEXT_DEFAULT)
	obs.obs_property_set_enabled(sourceProp, false)
	sourceProp = obs.obs_properties_add_text(props,"source_no","Source Id",obs.OBS_TEXT_DEFAULT)
	obs.obs_property_set_enabled(sourceProp, false)
	sourceProp=obs.obs_properties_add_bool(props,"desert_name","Desert Name")  
	obs.obs_property_set_enabled(sourceProp, false)
	obs.obs_property_set_visible(sourceProp, false)
	sourceProp=obs.obs_properties_add_bool(props,"assert_name","Use a descriptive name for this "..sourceType.." source")
	obs.obs_property_set_long_description(sourceProp,"Override the default source name to convey mute and volume effects")
	
	sourceProp = obs.obs_properties_add_list(props, "transition_action", "Take This Action" ,obs.OBS_COMBO_TYPE_LIST, obs.OBS_COMBO_FORMAT_STRING) 
	local msg="'Mute' and 'Unmute' are self-explanatory and instantaneous\n'Fade' gradually adjusts the volume to a desired rest level\n'Ramp' is the same as 'Fade', except silencing first\nSet adjusts volume without fading\n'Save' captures the volume and mute state for a subsequent restore\n'Restore' reverts to the most recently captured volume and mute state"
	obs.obs_property_set_long_description(sourceProp,msg)
	obs.obs_property_list_add_string(sourceProp, "Mute", "mute")
	obs.obs_property_list_add_string(sourceProp, "Unmute", "unmute")
	obs.obs_property_list_add_string(sourceProp, "Fade", "fade")
	obs.obs_property_list_add_string(sourceProp, "Ramp", "ramp")
	obs.obs_property_list_add_string(sourceProp, "Set", "set")
	obs.obs_property_list_add_string(sourceProp, "Save", "save")
	obs.obs_property_list_add_string(sourceProp, "Restore", "restore")
	obs.obs_property_set_modified_callback(sourceProp, onTransitionActionChange)
	
	sourceProp = obs.obs_properties_add_list(props, "audioSource", "Audio Source", obs.OBS_COMBO_TYPE_LIST, obs.OBS_COMBO_FORMAT_STRING)
	obs.obs_property_set_long_description(sourceProp,"Choose the Audio Source for the selected Action")
	local audioSourceList=getAudioSourceList()  
	table.sort(audioSourceList)
	obs.obs_property_list_add_string(sourceProp, all, all)
	for i, name in pairs(audioSourceList) do
		obs.obs_property_list_add_string(sourceProp, name, name)
	end

	sourceProp=obs.obs_properties_add_int_slider(props,"fader_level", "To Volume (-60 to 0)",faderMin,faderMax,1)
	obs.obs_property_set_long_description(sourceProp,"Fading and Ramping settle at this attenuation level")

	sourceProp = obs.obs_properties_add_list(props, "override_duration", "With Duration From", obs.OBS_COMBO_TYPE_LIST, obs.OBS_COMBO_FORMAT_STRING)
	obs.obs_property_set_long_description(sourceProp,"Choose where to get the transition duration")
	obs.obs_property_set_modified_callback(sourceProp, onDurationOverrideChange)
	obs.obs_property_list_add_string(sourceProp, "Scene Transitions Duration", "scene")
	obs.obs_property_list_add_string(sourceProp, "Custom Duration Slider", "slider")
	
	sourceProp=obs.obs_properties_add_int_slider(props,"fader_duration", "Custom Duration (milliseconds)",0,faderDurationMax*1000,10)
	obs.obs_property_set_long_description(sourceProp,"Fading, Ramping and Restoring progress over this duration instead of the Scene Transitions Duration")
	
	sourceProp=obs.obs_properties_add_list(props,"transition_side","When",obs.OBS_COMBO_TYPE_LIST, obs.OBS_COMBO_FORMAT_STRING)obs.obs_property_set_long_description(sourceProp,"Choose when to take the Action") 
	obs.obs_property_list_add_string(sourceProp, "Entering Program Scene", entering)
	obs.obs_property_list_add_string(sourceProp, "Leaving Program Scene", leaving)
	obs.obs_property_set_modified_callback(sourceProp, onTransitionSideChange)
	sourceProp= obs.obs_properties_add_bool(props,"auto_restore","Restore audio volume and mute state when transitioning to the next scene")
	obs.obs_property_set_long_description(sourceProp,"Re-establish the prior volume and mute state\nAn explicit 'Save' action is not required")
	return props	
end	
function source_defaults(settings)
	local sourceNo = getSourceNo (settings)	
	obs.obs_data_set_default_string(settings,"source_no", sourceNo)
	obs.obs_data_set_default_string(settings,"source_type",sourceType)
	obs.obs_data_set_default_bool(settings,"assert_name",true)
	obs.obs_data_set_default_bool(settings,"desert_name",false)
	obs.obs_data_set_default_string(settings,"transition_action","mute")
	obs.obs_data_set_default_string(settings,"transition_side",entering)
	obs.obs_data_set_default_string(settings,"audioSource",all)
	obs.obs_data_set_default_string(settings,"override_duration","scene")
	obs.obs_data_set_default_bool(settings, "auto_restore",false)
end

--helper functions
function perToDb(level)  --not used, "borrowed" from automute-duck.lua and preserved for reference purposes
  local def = level / 60.0
  local db = 0
  if (def >= 1.0) then 
	    db = 0.0
  elseif (def <= 0.0) then
		db = -100;
  else
		db = 6 - (102 * (1/(17 ^ def)))
  end
  return db
end
function fadToDb(level, precision)
  local vs=precision or vs
  local percent=(level+60)/60
  local db = 0
  if (level <= -60) then
		db = "-inf"			
  elseif (level < 0) then
		db = 6 - (102 * (1/(17 ^ percent))) --this is magic, can't pretend to understand
  end
--print ("fadToDb: fad="..level.." db="..db)
  return round(db,vs)
end
function fadToVol(level)
	local vs=precision or vs
	local db = fadToDb(level)
	local vol=dbToVol(db)
--print ("fadToVol: fad="..level.." db="..db.." vol="..vol)
	return round(vol,vs)
end
function fadToPer(level) 
	local percent=100*((level+60)/60)
	return round(percent,0)
end
function volToDb(vol, precision)
	local vs=precision or vs
	local db= math.log10(vol) * 20.0
--print ("volToDb: vol="..vol.." db="..db)
	return round(db,vs)
end
function dbToVol(db, precision)
	local vs=precision or vs
	local vol = math.pow(10, db / 20.0)
--print ("dbToVol: db="..db.." vol="..db)
	return round(vol,vs)
end

function round(number, precision)
	local precision = precision or 0  --default if precision omitted
	return tonumber(string.format("%."..precision.."f", number))
end
function toDec(value,length)
	local format = "%.0f"
	if length then format = "%."..tostring(length).."f" end
	return string.format(format,tostring(value))
end
function numerate (number, ...)  
	local arg={...}
	local txt=""
	local sep=""
	if number then txt="("..number..")" end 
	for i, msg in pairs(arg) do
		txt=txt..sep..msg
		sep = " "
	end
	return txt	
end
function toStr(value,length)
local spaced = string.rep(" ", length)..value
	return string.sub(spaced,#spaced-length) 
end

function cloneTable(table)
	local t={}
	for k,v in pairs(table) do
		t[k]=v
	end
	return t
end
function logError (msg)
	if errorPopupEnabled then
		error("Error: "..msg)
	else
		logIt("Error: "..msg)
	end
end
function trace(...)
--for temporary diagnostic traces..should never go into production
	print("Trace: ")
	local arg={...}
	local prefix = "   "
	for i, msg in pairs(arg) do	 
		print (prefix..msg) 
		prefix = "   "
	end
end
function elog (...) 
--log event info
	if not eventLogEnabled then return end 
	logIt (...)
end
function flog (fd, ...)
--log fader info
	if not faderLogEnabled then return end
	local prefix=numerate(fd.sourceNo,"["..fd.audioSourceName).."]: "
	logIt (prefix.. ... )
end
function logIt (...)
	local arg={...}
	local prefix = ">>>"
	if #arg == 0 then
		print (" ")
		return
	end
	for i, msg in pairs(arg) do	 
		print (prefix..msg) 
		prefix = "   "
	end
end

--event handlers
function onScriptLoaded()
	local scene=obs.obs_frontend_get_current_scene()
	obs.obs_source_release(scene)
	if scene then
		obs.timer_remove(onScriptLoaded)		
		onTransitionListChanged()
		elog(numerate(nil, sourceType.." Loaded..."))	
		onTransitionStart()
		return
	end
	elog (numerate(nil,"Waiting for current scene to load..."))
	loadAttempts=loadAttempts+1
	if loadAttempts > 10 then  --limit iterations
		obs.timer_remove(onScriptLoaded)
		onTransitionListChanged()
		logError(numerate(nil,"Excessive startup delay"))
	end
	--iterate 
end
function onTransitionStart()
	elog (numerate(nil," "))
	local enterSceneSource=obs.obs_frontend_get_current_scene() 
	leaveSceneName=enterSceneName
	local leaveSceneSource=obs.obs_get_source_by_name(leaveSceneName)  -- returns nil on first pass	
	enterSceneName=obs.obs_source_get_name(enterSceneSource)
	msg = "Scene transition from ["..leaveSceneName.."] to ["..enterSceneName.."] started"
	elog (numerate(nil,msg))
	
	if transitionsEnabled then
		transitionAudio(leaveSceneSource, enterSceneSource)
	end
	obs.obs_source_release(enterSceneSource)
	obs.obs_source_release(leaveSceneSource) 
end
function onTransitionStop()
	elog (numerate(nil,"Scene transition to ["..enterSceneName.."] completed"))
end
function onTransitionListChanged()
	--disconnect callbacks for all transitions
	for i, sh in ipairs (signalData) do
		obs.signal_handler_disconnect(sh,"transition_start",onTransitionStart)
		obs.signal_handler_disconnect(sh,"transition_stop",onTransitionStop)
		signalData[i] = nil
	end
	--connect callbacks for all transitions
	local transitionList =  obs.obs_frontend_get_transitions()
    for _, transitionSource in pairs(transitionList) do
		local sh = obs.obs_source_get_signal_handler(transitionSource)
        obs.signal_handler_connect(sh,"transition_start",onTransitionStart)
		obs.signal_handler_connect(sh,"transition_stop",onTransitionStop)
		table.insert(signalData,sh)
    end
	elog(numerate(nil,"Signal handlers connected for "..#signalData.." transition(s)"))
    obs.source_list_release(transitionList)
	transitionSignalsEnabled=true
end
function onTransitionSideChange(props, prop, settings)
	local restoreEnabled = false
	if obs.obs_data_get_string(settings, "transition_side") == entering then
		restoreEnabled = true
	end
	obs.obs_property_set_enabled (obs.obs_properties_get(props, "auto_restore"),restoreEnabled)
	return true
end
function onTransitionActionChange(props, prop, settings)
	local action = obs.obs_data_get_string(settings,"transition_action")
	local duration = obs.obs_data_get_string(settings,"override_duration")
	local isOverride = (action == "fade" or action == "ramp" or action == "restore")
	local isDuration =  isOverride and (duration == "slider")
	local isVolume = (action == "fade" or action == "ramp" or action == "set")
	
	obs.obs_property_set_enabled(obs.obs_properties_get(props, "fader_level"), isVolume)
	obs.obs_property_set_enabled(obs.obs_properties_get(props,"override_duration"), isOverride)
	obs.obs_property_set_enabled(obs.obs_properties_get(props, "fader_duration"), isDuration)
	return true	
end
function onDurationOverrideChange(props, prop, settings)
	local action = obs.obs_data_get_string(settings,"transition_action")
	local override = obs.obs_data_get_string(settings,"override_duration")
	local duration =  (action == "fade" or action == "ramp" or "restore") and override == "slider"
	obs.obs_property_set_enabled(obs.obs_properties_get(props,"fader_duration"),duration)
	return true	
end
function onFrontendEvent(event)
	if event == obs.OBS_FRONTEND_EVENT_TRANSITION_LIST_CHANGED then
		onTransitionListChanged()
	end
end
function onSourceSave(cd)
--[[
This callback is invoked for EVERY instance of the source, each time any of them is saved. It renames the sources to assure uniqueness and (optionally) reflect its property configuration. Since the rename invokes a save, care must be taken to avoid infinite recursion. Thus, the check to avoid the rename when the new name
and old name are the same.

But then...
OBS requires every source to have a unique name and appends a numeric suffix when a duplicate is saved. That creates a problem because the OBS "duplicate name accommodation" re-invokes this callback, causing an infinite loop. The employed solution is to append a unique sourceNo to the source name, avoiding the possibility of a duplicate name.
]]
	local source = obs.calldata_source(cd,"source")
	if source == nil then -- should never happen
		logError (numerate(nil,"Cannot save "..sourceType.." that no longer exists"))
		return
	end
	local sd=getAudioTransition(source)  
	local nameType=""
	if sd.assertName then 
		nameType="descriptive"
	else
	if sd.desertName then
		nameType="default" 
	else
		nameType="conventional"
	end
	end
	elog (numerate(sd.sourceNo,"Saving ["..sourceType.."("..sd.sourceNo..")]".." with ".. nameType.." name"))
	local audioSourceName = sd.audioSourceName
	if audioSourceName == "" then  --should not happen
		logError (numerate(sd.sourceNo,"Cannot resolve audio source name"))
		audioSourceName = "see Script Log"
	end
	
	local newName
	--desertName distinquishes a manual rename from and one that occurs intrinsically when assertName is turned off.
	--!may be a better way by adding a rename signal
	local settings = obs.obs_source_get_settings(source) 
	if sd.assertName then
		newName=getDescriptiveSourceName(sd)
		obs.obs_data_set_bool(settings,"desert_name", true)	
	else
		if sd.desertName then  --true briefly when assertName is turned off 
			newName = sourceType.."("..sd.sourceNo..")"	
			obs.obs_data_set_bool(settings,"desert_name", false)
		else
			newName = sd.sourceName   --for a manual rename
		end	
	end
	obs.obs_data_release(settings)
	
--this avoids an infinite loop caused by unconditionally invoking a save as the result of a rename
	if newName ~= sd.sourceName then
		elog (numerate( sd.sourceNo,"Renaming ["..sd.sourceName.."] to [" .. newName.."]"))
		obs.obs_source_set_name(source, newName)
	else
		elog(numerate( sd.sourceNo,"Rename completed"))
	end
end
function onSourceRename(cd, newName, oldName)
--!this signal useless, as newName and olrName are nil
--use save instead
--local source = obs.calldata_source(cd,"source")
--trace("rename", newName, oldName)
return
end
	
--diagnostic subroutines
function showDiagnostics()
	error (">>>") --not really an error, forces log to popup
	print("INSTALLATION:")  
	print (" "..sourceType .." " ..version.." " ..script_path())
	print ("AUDIO SOURCES:")
	local audioSourceList=getAudioSourceList()
	for i, audioSourceName in ipairs(audioSourceList) do
		local audioSource=obs.obs_get_source_by_name(audioSourceName)
		local volume=obs.obs_source_get_volume(audioSource)
		print (" ["..audioSourceName.."]: volume="..volume)
		obs.obs_source_release(audioSource)
	end
	print (string.upper(sourceType).. " SOURCES:")
	local sources=obs.obs_enum_sources()
	for i, source in ipairs(sources) do
		local settings = obs.obs_source_get_settings(source)
		if obs.obs_data_get_string(settings, "source_type") == sourceType then
			local sd=getAudioTransition(source)
			local actionTxt=""
			local durationTxt = " over ".. sd.faderDuration .." ms"
			if sd.faderOverrideDuration == "scene" then
				durationTxt= " over Scene Transition Duration"
			end
			if sd.transitionAction == "ramp" or sd.transitionAction == "fade" then
				actionTxt = sd.faderLevel.. " (volume="..sd.faderVolume..")".. durationTxt
			else
			if sd.transitionAction == "set" then
				actionTxt = sd.faderLevel.. " (volume="..sd.faderVolume..")"
			end
			end
			local msg=" Fader"..numerate(sd.sourceNo)..": "..sd.transitionAction
			msg=msg.." ["..sd.audioSourceName.."] to "..actionTxt.." on "..sd.transitionSide  --!show sceneName, but can be more than one
			if sd.autoRestore then msg=msg.." with revert" end
			print (msg)
		end
		obs.obs_data_release(settings)
	end
	obs.source_list_release(sources)
	print ("VOLUME CONVERSIONS:")
	showVolumeConversions()
end
function showVolumeConversions()
--for development, to list when tweaking attenuation conversion algorythm
	local fad = faderMax
	while fad >= faderMin do
		local fad2db=fadToDb(fad)
		local fad2vol=fadToVol(fad)
		local fad2per=fadToPer(fad)
		local db2vol=dbToVol(fad2db)
		local vol2db=volToDb(fad2vol)
		local msg=" Fader="..toStr(fad,3).." "..toStr(fad2db,12).. "dB "..toStr(fad2per,2).. "%".. " volume="..toDec(fad2vol,5)
		if not fad2vol == db2vol then msg=msg.." or "..db2vol  end
		print (msg)		
		fad = fad - 5
	end	
end
function showTable(table, level)  --for debugging
	local level=level or 0
	if level == 0 then 	print (tostring(table)) end	
	local indent= "   "..string.rep(" ",level*3)
	for k,v in pairs(table) do
		if type(v) == "table" then
			print (indent.."table: "..tostring(k))
			showTable(v,level+1)
		else
			print (indent..tostring(k).."=>"..tostring(v))
		end
	end
end
function showTime (msg)
	local time = os.date("*t")
	print(msg..time.hour .. ":" .. time.min .. ":" .. time.sec)
end

--worker subroutines
function registerPlugin()
	local outputFlags=obs.OBS_SOURCE_CUSTOM_DRAW  --need something
	local pluginDef = {}
	pluginDef.id = sourceType
	pluginDef.type =  obs.OBS_SOURCE_TYPE_INPUT 
	pluginDef.output_flags = outputFlags
	pluginDef.get_name = function()	return sourceType end 
	pluginDef.create = source_create
	pluginDef.get_properties = source_properties
	pluginDef.get_defaults = source_defaults
	obs.obs_register_source(pluginDef)
end
function getAudioSourceList()   --!add support for active only
	local audioSourceList={}
	local sources = obs.obs_enum_sources()
	--assemble all audio sources for dropdown list
	if sources == nil then   --this should never happen
		obs.source_list_release(sources)
		logError (numerate(nil,"No sources found in scene collection"))
		return
	end
	for _, source in ipairs(sources) do
		outputFlags=obs.obs_source_get_output_flags(source)
		if bit.band(outputFlags,obs.OBS_SOURCE_AUDIO) > 0 then
			table.insert(audioSourceList, obs.obs_source_get_name(source))
		end
	end
	obs.source_list_release(sources)
	return audioSourceList
end
function getSourceNo(settings)
--[[
This function returns the first available integer not already used as a sourceNo.
]]
--enumerate sourceNo's of this sourceType
	local sources = obs.obs_enum_sources()
	if sources == nil then  --this should never happen
		obs.source_list_release(sources) 
		logError (numerate(nil,"Cannot continue without sources in scene collection"))	
		return nil
	end
	local n = {}
	for i, iSource in ipairs(sources) do
		local iSourceType = obs.obs_source_get_id(iSource)
		if iSourceType == sourceType then
			local iSettings=obs.obs_source_get_settings(iSource)
			local iSourceNo  = obs.obs_data_get_string(iSettings, "source_no")
			if iSourceNo ~= "" then
				n[#n+1]=tonumber(iSourceNo)
			end
			obs.obs_data_release(iSettings)
		end
	end
	obs.source_list_release(sources)
--sort in numeric order then, then iterate over table until 
--an available number is found 
	table.sort(n)
	local lastId = 0
	local thisId = 0
	for i, iSourceNo in ipairs(n) do			
		thisId = tonumber(iSourceNo)
		if thisId > lastId + 1 then  --found a gap, use it
			return lastId+1   
		end
		lastId = thisId
	end
	return thisId+1  --no gaps, use next sequential number
end
function getAudioTransitions(sceneSource)
	if sceneSource==nil then return {} end
	local audioTransitions={}
	local sceneName=obs.obs_source_get_name(sceneSource)
	local scene=obs.obs_scene_from_source(sceneSource)
	local sceneItems=obs.obs_scene_enum_items(scene)
	for i, item in ipairs(sceneItems) do
		local itemSource = obs.obs_sceneitem_get_source(item) --need not release
		--[[
		As designed, inactive Audio Transition sources are not honored. However,
		inactive Audio Sources can be manipulated by active Audio Transition sources.		
		]] 
		local itemSourceName = obs.obs_source_get_name(itemSource)
		local itemSourceType = obs.obs_source_get_type(itemSource)		
		local sceneItemSource = obs.obs_get_source_by_name(itemSourceName)
		local settings=obs.obs_source_get_settings(itemSource)
		local itemType=obs.obs_data_get_string(settings, "source_type")
-- recognize active, audio transition sources  
		if obs.obs_source_active(sceneItemSource)
			and itemType== sourceType then
			local at=getAudioTransition(itemSource, sceneName)
			table.insert(audioTransitions,at)
		else
			--handle nested scenes
			if itemSourceType== obs.OBS_SOURCE_TYPE_SCENE then
				local nestedTransitions=getAudioTransitions(itemSource)
				if #nestedTransitions > 0 then
					elog(numerate(nil,#nestedTransitions.." nested Audio Transitions evaluated from ["..itemSourceName.."]"))
				end
				for _, nt in pairs(nestedTransitions) do
					table.insert(audioTransitions,nt)
				end
			end
		end
		obs.obs_data_release(settings)
		obs.obs_source_release(sceneItemSource)
	end
	obs.sceneitem_list_release(sceneItems)
	return audioTransitions
end
function getAudioTransition(source, sceneName)
	local settings = obs.obs_source_get_settings(source)
	local sd={}  -- source data
	local sceneSource=obs.obs_frontend_get_current_scene()
	sd.sceneName=sceneName
	sd.sourceName=obs.obs_source_get_name(source)
	sd.sourceNo=obs.obs_data_get_string(settings, "source_no")
	sd.sourceType=obs.obs_data_get_string(settings, "source_type")
	sd.audioSourceName=obs.obs_data_get_string(settings, "audioSource")
	sd.transitionSide=obs.obs_data_get_string(settings, "transition_side")
	sd.transitionAction=obs.obs_data_get_string(settings,"transition_action")
	sd.faderLevel = obs.obs_data_get_int(settings,"fader_level")
	sd.faderVolume = fadToVol(sd.faderLevel)
	sd.faderPercent= fadToPer(sd.faderLevel)
	sd.faderDb = fadToDb(sd.faderLevel,1)
	sd.faderDuration = obs.obs_data_get_int(settings,"fader_duration")
	sd.faderOverrideDuration = obs.obs_data_get_string(settings,"override_duration")
	sd.faderDelay=faderDelayBaseline+faderDelayOffset
	sd.assertName = obs.obs_data_get_bool(settings, "assert_name")
	sd.desertName = obs.obs_data_get_bool(settings, "desert_name")
	sd.autoRestore = obs.obs_data_get_bool(settings, "auto_restore") 
	obs.obs_data_release(settings)
	obs.obs_source_release(sceneSource)
	return sd
end
function getDescriptiveSourceName(sd)
	local midName = ""
	if sd.transitionAction == "ramp" or sd.transitionAction == "fade" or sd.transitionAction == "set" then
		midName = "<span style='color: red'> =>".. sd.faderLevel.." ("..sd.faderDb.." dB)</span>"
	end
	
	local newName = string.upper(string.sub(sd.transitionAction,1,1))..string.sub(sd.transitionAction,2).." <b>" .. sd.audioSourceName .. "</b>"..midName.." ".. sd.transitionSide .." ".. numerate(sd.sourceNo)
	if sd.autoRestore then newName=newName.." &#10226;" end   -- return symbol
	return newName
end
function saveAudioState(audioSourceName)
	local audioSource=obs.obs_get_source_by_name(audioSourceName)
	if audioSource == nil then
		logError (numerate(nil,"Cannot save audio state; Audio Source ["..audioSourceName.."] no longer exists"))
		return
	end 
	--instantiate audio state object and save it
	local as={}   
	as.restoreSourceName=audioSourceName
	as.restoreVolume = round(obs.obs_source_get_volume(audioSource),vs)
	as.restoreMuted = obs.obs_source_muted(audioSource)
	obs.obs_source_release(audioSource)
	audioData[audioSourceName] = as
	local restoreText="unmute"  
	if as.restoreMuted then restoreText = "mute" end  
	restoreText="("..restoreText.."->"..as.restoreVolume..")"
	elog (numerate(nil,"Saving audio state "..restoreText.." for ["..audioSourceName.."]"))
end
function saveLeaveData(sd, audioSourceName)
	elog (numerate(sd.sourceNo,"Accepting '"..sd.transitionAction .."' for ["..audioSourceName.."] when leaving [".. sd.sceneName.."]"))
	local ld=cloneTable(sd)
	ld.audioSourceName = audioSourceName
	table.insert(leaveData,ld)
end
function saveEnterData(sd, audioSourceName)
	elog (numerate(sd.sourceNo,"Accepting '"..sd.transitionAction.."' for ["..audioSourceName.."] when entering [".. sd.sceneName.. "]"))
	local ed=cloneTable(sd)
	ed.audioSourceName=audioSourceName
	table.insert(enterData,ed)
end

--core transition logic
function transitionAudio(leaveSceneSource, enterSceneSource)
--[[
This routine handles audio transition effects during a scene transition. At transition start,
it builds two queues of audio transitions: one for the scene leaving program view and the
other for the scene entering it. The queues preserve the order of sources in the respective
scene item lists. The leave queue is then processed first, followed by the enter queue.

An audio source is only allowed one fade or ramp transition per scene transition. As the queues
are processed, successive fade or ramp transitions for the same audio source replace their
predecessor.  In other words, transitions entering the program scene take precedence over those
leaving it and those higher in the scene item list take precedence over those lower down.
]]
	local transitionSource=obs.obs_frontend_get_current_transition()
	local transitionName = obs.obs_source_get_name(transitionSource)
	obs.obs_source_release(transitionSource)
	local transitionDuration = obs.obs_frontend_get_transition_duration()
	--message reports default transition, does not recognize a quick transition or a transition override 
	elog(numerate(nil,"Transitioning audio for Scene Transition "..transitionName.." "..transitionDuration .. "ms"))
	--Assemble Audio Transition sources in sceneItem (bottom up) order
	--clear out/ready transition queues 
	leaveData={} --clean slate for next transition
	enterData={}	
	local audioTransitions={}
	--iterate scene leaving program view, building queues of audio transitions
	audioTransitions=getAudioTransitions(leaveSceneSource)
	for _, sd in pairs(audioTransitions) do	
		if sd.transitionSide == leaving then
			local audioSourceList={sd.audioSourceName}
			if sd.audioSourceName == all then
				audioSourceList=getAudioSourceList()
			end
			for i, audioSourceName in pairs(audioSourceList) do
				saveLeaveData(sd, audioSourceName) 
			end
		end	
		--this identifies auto restores when leaving
		if sd.transitionSide == entering and sd.autoRestore then
			ld=	(sd)
			ld.transitionAction="restore"
			ld.transitionSide=leaving
			ld.faderDuration = 0
			ld.faderOverrideDuration = "zero"
			local audioSourceList = {sd.audioSourceName}
			if sd.audioSourceName == all then
				audioSourceList = getAudioSourceList()
			end
			for i, audioSourceName in pairs (audioSourceList) do
				saveLeaveData(ld, audioSourceName)
			end
		end 
	end
	
	--iterate scene entering program view, building queues of audio transitions
	audioTransitions=getAudioTransitions(enterSceneSource)
	for _, sd in pairs(audioTransitions) do
		if sd.transitionSide == entering then
			local audioSourceList= {sd.audioSourceName}
			if sd.audioSourceName == all then
				audioSourceList=getAudioSourceList()
			end
			for i, audioSourceName in pairs(audioSourceList) do
				saveEnterData(sd, audioSourceName)	
				if sd.autoRestore then   --generate save for auto restore
					local ed=cloneTable(sd)
					ed.audioSourceName=audioSourceName
					ed.transitionAction="save"
					ed.transitionSide=entering
					ed.autoRestore = false				
					saveLeaveData(ed, audioSourceName)
				end
			end		
		end
	end	
	
--handle leaving sources
	elog(numerate(nil,"Begin Audio Transitions leaving [" .. leaveSceneName.."]"))
	for leaveSourceName, ls in pairs(leaveData) do
		transitionAudioSource(ls)
	end
	elog(numerate(nil,"End of Audio Transitions leaving [" .. leaveSceneName.."]"))

--handle incoming sources
	elog(numerate(nil,"Begin Audio Transitions entering [" .. enterSceneName.."]"))	
	for enterSourceName, es in pairs(enterData) do
		transitionAudioSource(es)	
	end
	elog(numerate(nil,"End of Audio Transition entering [" .. enterSceneName.."]"))
end
function transitionAudioSource(as)  
	local audioSource = obs.obs_get_source_by_name(as.audioSourceName)
	if audioSource == nil then
		logError (numerate(as.sourceNo,"Audio Source ["..as.audioSourceName.."] does not exist; '"..as.transitionAction.."' transition failed"))
		return
	end
	if as.transitionAction == "restore"
		and audioData[as.audioSourceName]==nil then
		obs.obs_source_release(audioSource)
		logError (numerate(as.sourceNo,"No saved audio state for ["..as.audioSourceName.."]; '"..as.transitionAction.. "' transition failed"))
		return
	end
	if as.faderOverrideDuration == "scene" then
		as.faderDuration = obs.obs_frontend_get_transition_duration() 
	end	
	if as.faderDuration == 0 then  --provision to make duration 0 instantaneous
		as.faderDelay = 1
		as.faderSteps = 1
	end
		
	local isMuted = obs.obs_source_muted(audioSource)
	local transitionMsg = as.transitionAction
	if as.transitionAction == "fade" or as.transitionAction=="ramp" or as.transitionAction=="set" then
		transitionMsg=transitionMsg.."(->"..as.faderVolume..")"
	else
	if as.transitionAction == "restore" then
		local ad=audioData[as.audioSourceName] 	
		local restoreText="unmute"  
		if ad.restoreMuted then restoreText = "mute" end  
		transitionMsg=transitionMsg.."("..restoreText.."->"..ad.restoreVolume..""
	end
	end
	elog (numerate(as.sourceNo,"Applying '"..transitionMsg.."' to ["..as.audioSourceName.."] when "..as.transitionSide.. " [".. as.sceneName.. "]"))
	
	if as.transitionAction=="mute" then
		if not isMuted then	
			obs.obs_source_set_muted(audioSource,true)  -- mute/unmute
		end
	else
	if as.transitionAction=="unmute" then
		if isMuted then	
			obs.obs_source_set_muted(audioSource,false)  -- mute/unmute
		end
	else
	if as.transitionAction=="fade" then   
		if isMuted then
			obs.obs_source_set_muted(audioSource,false)  --unmute
		end
		fadeAudio(audioSource, as) 
	else
	if as.transitionAction=="ramp" then   
		obs.obs_source_set_volume(audioSource,0)  --start at silence
		if isMuted then
			obs.obs_source_set_muted(audioSource,false)  --unmute
		end
		fadeAudio(audioSource, as)
	else
	if as.transitionAction=="set" then	
		obs.obs_source_set_volume(audioSource,as.faderVolume)
		if isMuted then  
			obs.obs_source_set_muted(audioSource,false)  --unmute
		end
	else
	if as.transitionAction=="restore" then
		local ad=audioData[as.audioSourceName]  
		obs.obs_source_set_muted(audioSource,ad.restoreMuted)
		as.faderVolume=ad.restoreVolume
		fadeAudio(audioSource, as)
	else
	if as.transitionAction=="save" then
		saveAudioState(as.audioSourceName)
	else
		logError (numerate(as.sourceNo,"'"..as.transitionAction.."' is not a supported Audio Transition action"))
	end
	end
	end
	end
	end
	end
	end
	obs.obs_source_release(audioSource)
end
function fadeAudio(audioSource, as)  
-- audio fader, controls gradual volume changes
-- construct fader data structure
	local fd={}
	fd.sourceNo=as.sourceNo
	fd.audioSourceName=as.audioSourceName
	fd.transitionAction=as.transitionAction
	fd.startVolume=obs.obs_source_get_volume(audioSource) 
	fd.lastVolume=fd.startVolume
	fd.targetVolume=as.faderVolume
	fd.faderLevel=as.faderLevel
	fd.faderDelay=math.max(1,as.faderDelay)
	fd.faderDuration=as.faderDuration
	fd.faderSteps=math.max(1,math.floor(fd.faderDuration/fd.faderDelay))
	fd.stepsRemaining=fd.faderSteps	
	
	local msg="Initiating fader("..fd.sourceNo..") for ["..fd.audioSourceName.."] from "..fd.startVolume.." to "..fd.targetVolume.. " in "..fd.faderSteps.." steps with "..fd.faderDelay.."ms delay"
	elog (numerate(as.sourceNo,msg))
	if faderData[as.audioSourceName] ~= nil then
		local fx=faderData[fd.audioSourceName]
		elog (numerate(fd.sourceNo,"Fader("..fd.sourceNo ..") succeeding fader("..fx.sourceNo..") for [".. fd.audioSourceName.. "]"))
		faderData[fd.audioSourceName]=fd        --replace/restart fader progress
		return
	end
	if fd.targetVolume == fd.startVolume then  --nothing to do
		faderData[as.audioSourceName]=nil
		elog (numerate(fd.sourceNo,"Fader("..fd.sourceNo..") dismissed, already at "..toDec(fd.targetVolume)))
		return   
	end
	local faderCallback=function()
		local fader=faderData[as.audioSourceName]
		if fader == nil then 
			logError(numerate(fd.sourceNo,"Fader for ["..as.audioSourceName.."] no longer exists"))		
			obs.remove_current_callback()  
			return
		end
		if fader.stepsRemaining == 0 then
			obs.remove_current_callback()
			faderData[as.audioSourceName]=nil
			flog (fader,"Fader("..fader.sourceNo..") steps exhausted")
			return
		end
		
		local audioSource = obs.obs_get_source_by_name(fader.audioSourceName)	
		if audioSource == nil then
			obs.remove_current_callback()
			logError(numerate(fader.sourceNo,"Fader abandoned; Audio Source ["..fader.audioSourceName.."] no longer exists"))
			return
		end

		local currentVolume = obs.obs_source_get_volume(audioSource) 
		local stepVol=(fader.targetVolume-currentVolume)/fader.stepsRemaining
		local nextVolume = currentVolume+stepVol
		local msg="countdown="..fader.stepsRemaining.." volume="..toDec(currentVolume,vs).."->"..toDec(nextVolume,vs).." step="..toDec(stepVol,vs)
		flog (fader,msg)

		if currentVolume ~= fader.lastVolume then
			flog(fader,"Fader("..fader.sourceNo..") abandoned due to conflicting volume change")
			obs.obs_source_release(audioSource)
			obs.remove_current_callback()
			faderData[as.audioSourceName]=nil
			return
		end		
	
		if fader.startVolume < fader.targetVolume then    --increasing volume
 			if nextVolume >= fader.targetVolume then  --too much
				flog(fader,"Fader("..fader.sourceNo..") has increased to desired rest volume") 
				nextVolume = fader.targetVolume
			end
		else
		if fader.startVolume > fader.targetVolume then   --decreasing volumes
			if nextVolume <= fader.targetVolume then  --too much
				flog(fader,"Fader("..fader.sourceNo..") has decreased to desired rest volume")
				nextVolume=fader.targetVolume
			end
		end
		end

		obs.obs_source_set_volume(audioSource,nextVolume)
		nextVolume=obs.obs_source_get_volume(audioSource) 
		faderData[fader.audioSourceName].lastVolume=nextVolume
		faderData[fader.audioSourceName].stepsRemaining = fader.stepsRemaining - 1 
		obs.obs_source_release(audioSource)
		if round(nextVolume,vs) == round(fader.targetVolume,vs) then
			flog (fader,"Fader("..fader.sourceNo..") has reached volume="..round(nextVolume,vs))	
		else
		if nextVolume == currentVolume then
			flog(fader,"Fader("..fader.sourceNo..") abandoned for lack of progress")
		else
			return    --iterate
		end
		end
		flog (fader,"Fader("..fader.sourceNo..") finishing at "..fader.faderLevel.. "("..volToDb(nextVolume,1).." dB)")
		faderData[fader.audioSourceName]=nil	
		obs.remove_current_callback()	
	end  --end fader callback
	
	faderData[as.audioSourceName]=fd
	obs.timer_add(faderCallback,fd.faderDelay)
end	
